/*******************************************************************************
 * Copyright 2015-2016 Radosław Piliszek
 *
 * This file is part of WSX2MIZ.
 *
 * WSX2MIZ is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * WSX2MIZ is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with WSX2MIZ.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package net.piliszek.mizar.wsx2miz;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.util.HashMap;
import java.util.Map;
import java.util.TreeSet;

import javax.xml.namespace.QName;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;

import org.apache.commons.lang3.tuple.ImmutablePair;

public class FRM {

	private static XMLInputFactory xif = XMLInputFactory.newInstance();

	private static Map<Integer, Integer> functorPriorities;

	private static Map<Integer, TreeSet<Integer>> modeArgumentCounts;
	private static Map<Integer, TreeSet<ImmutablePair<Integer, Integer>>> functorArgumentCounts;
	private static Map<Integer, TreeSet<Integer>> predicateRightArgumentCounts;
	private static Map<Integer, TreeSet<Integer>> predicateLeftArgumentCounts;

	private static QName kindQName = new QName("kind");
	private static QName symbolnrQName = new QName("symbolnr");
	private static QName valueQName = new QName("value");
	private static QName argnrQName = new QName("argnr");
	private static QName leftargnrQName = new QName("leftargnr");

	public static void load(File f)
			throws FileNotFoundException, XMLStreamException {
		functorPriorities = new HashMap<>();
		modeArgumentCounts = new HashMap<>();
		functorArgumentCounts = new HashMap<>();
		predicateRightArgumentCounts = new HashMap<>();
		predicateLeftArgumentCounts = new HashMap<>();

		XMLEventReader xer = xif.createXMLEventReader(
				new BufferedInputStream(new FileInputStream(f)));

		while (xer.hasNext()) {
			XMLEvent xe = xer.nextEvent();
			if (xe.isStartElement()) {
				StartElement se = xe.asStartElement();
				String elemName = se.getName().getLocalPart();

				if (elemName == "Priority") {
					String kind = se.getAttributeByName(kindQName).getValue();
					int symbolnr = Integer.parseInt(
							se.getAttributeByName(symbolnrQName).getValue());
					int value = Integer.parseInt(
							se.getAttributeByName(valueQName).getValue());
					if (kind.equals("O")) { // operation functor (operator)
						functorPriorities.put(symbolnr, value);
					}
				} else if (elemName == "Format") {
					String kind = se.getAttributeByName(kindQName).getValue();
					int symbolnr = Integer.parseInt(
							se.getAttributeByName(symbolnrQName).getValue());
					int argnr = Integer.parseInt(
							se.getAttributeByName(argnrQName).getValue());
					if (kind.equals("M")) { // mode
						registerModeFormat(symbolnr, argnr);
					} else if (kind.equals("O")) { // operation functor (operator)
						int leftargnr = Integer.parseInt(
								se.getAttributeByName(leftargnrQName)
										.getValue());
						registerFunctorFormat(symbolnr, leftargnr, argnr - leftargnr);
					} else if (kind.equals("R")) { // predicate
						int leftargnr = Integer.parseInt(
								se.getAttributeByName(leftargnrQName)
										.getValue());
						TreeSet<Integer> rightCounts = predicateRightArgumentCounts
								.get(symbolnr);
						if (null == rightCounts) {
							rightCounts = new TreeSet<>();
							predicateRightArgumentCounts.put(symbolnr,
									rightCounts);
						}
						rightCounts.add(argnr - leftargnr);
						TreeSet<Integer> leftCounts = predicateLeftArgumentCounts
								.get(symbolnr);
						if (null == leftCounts) {
							leftCounts = new TreeSet<>();
							predicateLeftArgumentCounts.put(symbolnr,
									rightCounts);
						}
						leftCounts.add(leftargnr);
					}
				}
			}
		}
	}

	public static int getFunctorPriority(int nr) {
		return functorPriorities.get(nr);
	}

	public static void registerModeFormat(int nr, int argnr) {
		TreeSet<Integer> counts = modeArgumentCounts.get(nr);
		if (null == counts) {
			counts = new TreeSet<>();
			modeArgumentCounts.put(nr, counts);
		}
		counts.add(argnr);
	}

	public static void registerFunctorFormat(int nr, int lArgCount, int rArgCount) {
		TreeSet<ImmutablePair<Integer, Integer>> counts = functorArgumentCounts.get(nr);
		if (null == counts) {
			counts = new TreeSet<>();
			functorArgumentCounts.put(nr, counts);
		}
		counts.add(ImmutablePair.of(lArgCount, rArgCount));
	}

	public static boolean isModeAmbiguous(int nr) {
		TreeSet<Integer> counts = modeArgumentCounts.get(nr);
		return counts.size() > 2 || (counts.size() == 2 && !counts.contains(0));
	}

	public static boolean canFunctorTake(int nr, int lArgCount, int rArgCount) {
		return functorArgumentCounts.get(nr).contains(ImmutablePair.of(lArgCount, rArgCount));
	}

}
