/*******************************************************************************
 * Copyright 2015-2016 Radosław Piliszek
 *
 * This file is part of WSX2MIZ.
 *
 * WSX2MIZ is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * WSX2MIZ is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with WSX2MIZ.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package net.piliszek.mizar.wsx2miz;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.NoSuchFileException;
import java.nio.file.NotDirectoryException;
import java.nio.file.Path;
import java.nio.file.Paths;

import javax.xml.XMLConstants;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.SchemaOutputResolver;
import javax.xml.bind.UnmarshalException;
import javax.xml.bind.Unmarshaller;
import javax.xml.bind.ValidationEvent;
import javax.xml.bind.ValidationEventHandler;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.stream.XMLStreamException;
import javax.xml.transform.Result;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.validation.SchemaFactory;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.DefaultParser;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.eclipse.persistence.exceptions.XMLMarshalException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import net.piliszek.mizar.wsx2miz.schema.TextProper;

public class Main {

	private static Unmarshaller unmarshaller;
	private static DocumentBuilder documentBuilder;
	private static Transformer transformer;
	private static Path outputDir;

	public static class MyValidationEventHandler
			implements ValidationEventHandler {

		public boolean handleEvent(ValidationEvent event) {
			if (event.getMessage().startsWith("cvc-complex-type.2.1:")) {
				return true;
			}
			return false;
		}

	}

	public static void preprocess(Document d) {
		Element e = d.getDocumentElement();
		e.setAttribute("xmlns:xsi",
				"http://www.w3.org/2001/XMLSchema-instance");

		preprocess(d, e);
	}

	public static void preprocess(Document d, Node n) {
		if (n.getNodeType() == Node.ELEMENT_NODE) {
			Element e = (Element) n;

			if (e.hasAttribute("kind")) {
				if (e.hasAttribute("shape")) {
					d.renameNode(e, null,
							e.getNodeName() + "-" + e.getAttribute("kind")
									+ "-" + e.getAttribute("shape"));
					e.removeAttribute("shape");
				} else {
					d.renameNode(e, null,
							e.getNodeName() + "-" + e.getAttribute("kind"));
				}
				e.removeAttribute("kind");
			}
		}

		NodeList children = n.getChildNodes();

		for (int i = 0; i < children.getLength(); i++) {
			preprocess(d, children.item(i));
		}
	}

	public static void mizarify(String arg) throws IOException,
			TransformerException, JAXBException {
		File wsxFile = new File(arg + ".wsx");
		File frmFile = new File(arg + ".frm");
		File dcxFile = new File(arg + ".dcx");

		try {
			FRM.load(frmFile);
		} catch (XMLStreamException e) {
			System.err.printf("%s is not a valid XML file.%n", frmFile);
			return;
		} catch (FileNotFoundException e) {
			System.err.printf("%s does not exist.%n", frmFile);
			return;
		}

		try {
			DCX.load(dcxFile);
		} catch (XMLStreamException e) {
			System.err.printf("%s is not a valid XML file.%n", dcxFile);
			return;
		} catch (FileNotFoundException e) {
			System.err.printf("%s does not exist.%n", dcxFile);
			return;
		}

		Document d;
		try {
			d = documentBuilder.parse(wsxFile);
		} catch (SAXException e) {
			System.err.printf("%s is not a valid XML file.%n", wsxFile);
			return;
		} catch (FileNotFoundException e) {
			System.err.printf("%s does not exist.%n", wsxFile);
			return;
		}

		preprocess(d);

		File tmpf = File.createTempFile("wsx", ".xml");

		try {
			transformer.transform(new DOMSource(d), new StreamResult(tmpf));

			try {
				System.out.printf("Processing %s%n",
						wsxFile.getCanonicalPath());
				TextProper tp = (TextProper) unmarshaller
						.unmarshal(tmpf);

				String filename = tp.getArticleId().toLowerCase() + ".miz";
				Path output = outputDir.resolve(filename);
				System.out.printf("Saving %s%n", output.toString());

				MizWriter w = new MizWriter(
						new BufferedWriter(new FileWriter(output.toFile())));

				w.write(tp);

				w.close();
			} catch (UnmarshalException e) {
				System.err.printf("Parse error in %s%n", wsxFile);
				try {
					XMLMarshalException mEx = (XMLMarshalException) e
							.getLinkedException();
					SAXParseException pEx = (SAXParseException) mEx
							.getInternalException();
					System.err.printf("Line: %d%n",
							pEx.getLineNumber() + 1);
					System.err.println(pEx.getMessage());
				} catch (ClassCastException cce) {
					throw e;
				}
			}
		} finally {
			tmpf.delete();
		}
	}

	public static void main(String[] args)
			throws IOException, JAXBException, ParserConfigurationException,
			SAXException, TransformerException, XMLStreamException {
		try {
			Options options = new Options();
			options.addOption("o", true, "Path to output directory.");
			CommandLineParser parser = new DefaultParser();
			CommandLine line = parser.parse(options, args);
			args = line.getArgs();
			outputDir = Paths.get(line.getOptionValue("o", System.getProperty("java.io.tmpdir")));
		} catch (ParseException e) {
			System.err.println(e.getMessage());
			return;
		}

		if (!Files.isDirectory(outputDir)) {
			System.err.println("The output path is not a directory path.");
			return;
		}

		if (args.length == 0) {
			String s = String.join(System.getProperty("line.separator"), "WSX2MIZ", "",
					"Copyright 2015-2016 Radosław Piliszek", "",
					"This program is free software: you can redistribute it and/or modify",
					"it under the terms of the GNU Affero General Public License as published by",
					"the Free Software Foundation, either version 3 of the License, or",
					"(at your option) any later version.", "",
					"This program is distributed in the hope that it will be useful,",
					"but WITHOUT ANY WARRANTY; without even the implied warranty of",
					"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the",
					"GNU Affero General Public License for more details.", "",
					"You should have received a copy of the GNU Affero General Public License",
					"along with this program.  If not, see <http://www.gnu.org/licenses/>");
			System.out.println(s);
			System.out.println();
			System.err.println("At least one argument is required.");
			return;
		}

		File schemaFile = File.createTempFile("wsx", ".xsd");
		schemaFile.deleteOnExit();

		JAXBContext jaxbContext = JAXBContext.newInstance(TextProper.class);
		System.out.printf("Saving schema to %s%n",
				schemaFile.getCanonicalPath());
		jaxbContext.generateSchema(new SchemaOutputResolver() {

			@Override
			public Result createOutput(String namespaceUri,
					String suggestedFileName) throws IOException {
				StreamResult result = new StreamResult(schemaFile);
				return result;
			}

		});

		unmarshaller = jaxbContext.createUnmarshaller();
		SchemaFactory sf = SchemaFactory
				.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
		unmarshaller.setSchema(sf.newSchema(schemaFile));
		unmarshaller.setEventHandler(new MyValidationEventHandler());

		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		dbf.setNamespaceAware(true);
		documentBuilder = dbf.newDocumentBuilder();
		documentBuilder.setErrorHandler(null);

		TransformerFactory tf = TransformerFactory.newInstance();
		transformer = tf.newTransformer();

		for (String arg : args) {
			try (DirectoryStream<Path> stream = Files
					.newDirectoryStream(Paths.get(arg), "*.wsx")) {
				for (Path entry : stream) {
					String path = entry.toString();
					mizarify(path.substring(0, path.lastIndexOf('.')));
				}
			} catch (NoSuchFileException | NotDirectoryException e) {
				mizarify(arg);
			}
		}
	}

}
