/*******************************************************************************
 * Copyright 2015-2016 Radosław Piliszek
 *
 * This file is part of WSX2MIZ.
 *
 * WSX2MIZ is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * WSX2MIZ is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with WSX2MIZ.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package net.piliszek.mizar.wsx2miz;

import java.io.Closeable;
import java.io.Flushable;
import java.io.IOException;
import java.io.Writer;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import net.piliszek.mizar.wsx2miz.schema.Example;
import net.piliszek.mizar.wsx2miz.schema.ExplicitlyQualifiedSegment;
import net.piliszek.mizar.wsx2miz.schema.ImplicitlyQualifiedSegment;
import net.piliszek.mizar.wsx2miz.schema.ItemCaseBlock;
import net.piliszek.mizar.wsx2miz.schema.ItemExemplification;
import net.piliszek.mizar.wsx2miz.schema.ItemPerCases;
import net.piliszek.mizar.wsx2miz.schema.Justification;
import net.piliszek.mizar.wsx2miz.schema.Proposition;
import net.piliszek.mizar.wsx2miz.schema.QualifiedSegment;
import net.piliszek.mizar.wsx2miz.schema.ReasoningItem;
import net.piliszek.mizar.wsx2miz.schema.SimpleJustification;
import net.piliszek.mizar.wsx2miz.schema.Term;
import net.piliszek.mizar.wsx2miz.schema.helpers.WSXElement;

public class MizWriter implements Closeable, Flushable {

	private static final int DEFAULT_INDENT = 2;
	private static final String NEW_LINE = "\r\n";
	private static final int MAX_LENGTH = 79;

	private int currentIndent = 0;
	private int currentLength = 0;

	private boolean writeSpace = false;

	private LinkedList<StringBuilder> buffers = new LinkedList<>();

	private Writer out;

	public MizWriter(Writer out) {
		this.out = out;
	}

	public void flush() throws IOException {
		out.flush();
	}

	public void close() throws IOException {
		out.close();
	}

	public void increaseIndent() {
		currentIndent += DEFAULT_INDENT;
	}

	public void decreaseIndent() {
		currentIndent -= DEFAULT_INDENT;
	}

	public void newLine() throws IOException {
		if (buffers.size() > 0) {
			throw new RuntimeException("Unfinished subexp.");
		}
		writeNewLine();
		writeSpace = false;
	}

	private void writeNewLine() throws IOException {
		out.write(NEW_LINE);
		currentLength = 0;
	}

	public void beginSubExp() throws IOException {
		StringBuilder buffer = buffers.peekLast();

		if (null != buffer && buffer.length() == 0) {
			throw new RuntimeException("Buffer state is wrong.");
		}

		buffers.addLast(new StringBuilder());
	}

	public void endSubExp() throws IOException {
		StringBuilder buffer = buffers.pollLast();

		if (null != buffer) {
			if (buffer.length() == 0) {
				throw new RuntimeException("Buffer state is wrong.");
			}

			StringBuilder pBuffer = buffers.peekLast();

			if (null != pBuffer) {
				pBuffer.append(buffer);
			} else {
				out.write(buffer.toString());
			}
		}
	}

	public void ensureSpace() throws IOException {
		if (writeSpace) {
			throw new RuntimeException("Space state is wrong.");
		}

		writeSpace = true;
	}

	public void write(String s) throws IOException {
		if ("".equals(s)) {
			return;
		}

		if (s.contains("\n") || s.contains("\r")) {
			throw new RuntimeException("New line in token.");
		}

		if (s.startsWith(" ") || s.endsWith(" ")) {
			throw new RuntimeException(
					"Whitespace in token as prefix or suffix.");
		}

		writeToken(s);
	}

	private void writeToken(String s) throws IOException {
		if (currentLength + (writeSpace ? 1 : 0) + s.length() > MAX_LENGTH) {
			if (!buffers.isEmpty()) {
				StringBuilder buffer = buffers.pollFirst();
				String str = buffer.toString().replaceFirst("^\\s+", "");
				if (!"".equals(str)) {
					writeNewLine();
					writeIndent();
					out.write(str);
					currentLength += str.length();
					for (StringBuilder sb : buffers) {
						currentLength += sb.length();
					}
					writeToken(s);
					return;
				}
			}

			writeNewLine();
			writeIndent();
			writeSpace = false;
		}

		StringBuilder buffer = buffers.peekLast();

		if (writeSpace) {
			if (null != buffer) {
				buffer.append(' ');
			} else {
				out.write(' ');
			}
			currentLength += 1;
			writeSpace = false;
		}

		if (null != buffer) {
			buffer.append(s);
		} else {
			out.write(s);
		}
		currentLength += s.length();
	}

	public void writeIndent() throws IOException {
		for (int i = 0; i < currentIndent; i++) {
			out.write(' ');
		}
		currentLength += currentIndent;
	}

	public void write(WSXElement e) throws IOException {
		e.write(this);
	}

	public static void optimizeReasoning(List<ReasoningItem> reasoningItems) {
		Iterator<ReasoningItem> it = reasoningItems.iterator();
		ItemExemplification lastExemplification = null;
		while (it.hasNext()) {
			ReasoningItem ri = it.next();
			try {
				ItemExemplification ex = (ItemExemplification) ri;
				if (null != lastExemplification) {
					lastExemplification.addExample(ex.getFirstExample());
					it.remove();
				} else {
					lastExemplification = ex;
				}
			} catch (ClassCastException e) {
				lastExemplification = null;
			}
		}
	}

	public void writeReasoning(
			List<ReasoningItem> reasoningItems,
			ItemPerCases perCases,
			List<ItemCaseBlock> caseBlocks) throws IOException {
		increaseIndent();
		if (reasoningItems != null) {
			optimizeReasoning(reasoningItems);
			for (ReasoningItem ri : reasoningItems) {
				write(ri);
			}
		}
		if (perCases != null) {
			write(perCases);
			for (ItemCaseBlock cb : caseBlocks) {
				write(cb);
			}
		}
		decreaseIndent();
	}

	public void writeJustification(SimpleJustification justification)
			throws IOException {
		if (justification.hasReferences()) {
			ensureSpace();
			write(justification);
		}
	}

	public void writeJustification(Justification justification)
			throws IOException {
		try {
			writeJustification((SimpleJustification) justification);
		} catch (ClassCastException e) {
			newLine();
			write(justification);
		}
	}

	public void writeQualifiedSegments(List<QualifiedSegment> qualifiedSegments,
			String be) throws IOException {
		Iterator<QualifiedSegment> it = qualifiedSegments.iterator();
		while (true) {
			QualifiedSegment s = it.next();
			if (it.hasNext()) {
				s.write(this, be,
						WSXElement.UNSAFE_RIGHT_DANGLING_COMMA);
				write(",");
				ensureSpace();
			} else {
				s.write(this, be, 0);
				break;
			}
		}
	}

	public void writeImplicitlyQualifiedSegments(
			List<ImplicitlyQualifiedSegment> qualifiedSegments)
					throws IOException {
		String d = "";
		for (ImplicitlyQualifiedSegment iqs : qualifiedSegments) {
			write(d);
			write(iqs);
			d = ",";
		}
	}

	public void writeExplicitlyQualifiedSegments(
			List<ExplicitlyQualifiedSegment> qualifiedSegments,
			String be, int flags) throws IOException {
		Iterator<ExplicitlyQualifiedSegment> it = qualifiedSegments.iterator();
		while (true) {
			QualifiedSegment s = it.next();
			if (it.hasNext()) {
				s.write(this, be,
						WSXElement.UNSAFE_RIGHT_DANGLING_COMMA);
				write(",");
				ensureSpace();
			} else {
				s.write(this, be, flags);
				break;
			}
		}
	}

	public void writeExamples(Iterable<? extends Example> examples) throws IOException {
		Iterator<? extends Example> it = examples.iterator();
		while (true) {
			Example e = it.next();
			if (it.hasNext()) {
				beginSubExp();
				e.write(this, WSXElement.UNSAFE_RIGHT_DANGLING_COMMA);
				endSubExp();
				write(",");
				ensureSpace();
			} else {
				beginSubExp();
				e.write(this, 0);
				endSubExp();
				break;
			}
		}
	}

	public void writeTerms(List<Term> terms, int flags) throws IOException {
		int currentFlags = WSXElement.UNSAFE_LEFT_MASK & flags;
		Iterator<Term> it = terms.iterator();
		while (true) {
			Term t = it.next();
			if (it.hasNext()) {
				t.write(this,
						WSXElement.UNSAFE_RIGHT_DANGLING_COMMA | currentFlags);
				write(",");
				ensureSpace();
			} else {
				t.write(this,
						(WSXElement.UNSAFE_RIGHT_MASK & flags) | currentFlags);
				break;
			}
			currentFlags = 0;
		}
	}

	public void writePropositions(List<Proposition> propositions)
			throws IOException {
		Iterator<Proposition> it = propositions.iterator();
		while (true) {
			Proposition p = it.next();
			writeIndent();
			p.write(this);
			if (it.hasNext()) {
				ensureSpace();
				write("and");
				newLine();
			} else {
				break;
			}
		}
	}

}
