/*******************************************************************************
 * Copyright 2015-2016 Radosław Piliszek
 *
 * This file is part of WSX2MIZ.
 *
 * WSX2MIZ is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * WSX2MIZ is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with WSX2MIZ.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package net.piliszek.mizar.wsx2miz.schema;

import java.io.IOException;
import java.util.List;

import javax.xml.bind.Unmarshaller;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlElementRef;
import javax.xml.bind.annotation.XmlRootElement;

import net.piliszek.mizar.wsx2miz.DCX;
import net.piliszek.mizar.wsx2miz.FRM;
import net.piliszek.mizar.wsx2miz.MizWriter;

@XmlRootElement(name = "Infix-Term")
public class InfixTerm extends Term {

	@XmlAttribute(name = "nr", required = true)
	protected int nr;
	@XmlAttribute(name = "spelling", required = true)
	protected String spelling;

	@XmlElementRef(required = true)
	protected List<Arguments> arguments;

	private Arguments argumentsL;
	private Arguments argumentsR;
	private int priority;

	protected void afterUnmarshal(Unmarshaller unmarshaller, Object parent) {
		argumentsL = arguments.get(0);
		argumentsR = arguments.get(1);
		priority = FRM.getFunctorPriority(nr);
	}

	@Override
	public boolean hasPriorityLessThan(int priority) {
		return this.priority < priority;
	}

	@Override
	public boolean hasPriorityLessThanOrEqual(int priority) {
		return this.priority <= priority;
	}

	private void writeLeftArguments(MizWriter w, int flags, InfixTerm left) throws IOException {
		List<Term> terms = argumentsL.terms;
		if (terms.size() > 1) {
			w.write("(");
			w.writeTerms(terms, UNSAFE_LEFT_PAREN | UNSAFE_RIGHT_PAREN);
			w.write(")");
		} else {
			Term t = terms.get(0);
			if (t.hasPriorityLessThan(priority)) {
				if ((flags & UNSAFE_LEFT_FUNCTOR) != 0) {
					t.write(w, UNSAFE_LEFT_WEAK_FUNCTOR
							| UNSAFE_RIGHT_WEAK_FUNCTOR | flags, left, this);
				} else {
					t.write(w, UNSAFE_RIGHT_WEAK_FUNCTOR | flags, left, this);
				}
			} else {
				t.write(w, UNSAFE_RIGHT_FUNCTOR | UNSAFE_RIGHT_DANGLING_TERM
						| flags, left, this);
			}
		}
	}

	private void writeRightArguments(MizWriter w, int flags, InfixTerm right)
			throws IOException {
		List<Term> terms = argumentsR.terms;
		if (terms.size() > 1) {
			w.write("(");
			w.writeTerms(terms, UNSAFE_LEFT_PAREN | UNSAFE_RIGHT_PAREN);
			w.write(")");
		} else {
			Term t = terms.get(0);
			if (t.hasPriorityLessThanOrEqual(priority)) {
				if ((flags & UNSAFE_RIGHT_FUNCTOR) != 0) {
					t.write(w, UNSAFE_RIGHT_WEAK_FUNCTOR
							| UNSAFE_LEFT_WEAK_FUNCTOR | flags, this, right);
				} else {
					t.write(w, UNSAFE_LEFT_WEAK_FUNCTOR | flags, this, right);
				}
			} else {
				t.write(w, UNSAFE_LEFT_FUNCTOR | UNSAFE_LEFT_DANGLING_TERM
						| flags, this, right);
			}
		}
	}

	private void writeIt(MizWriter w, int flags, InfixTerm left, InfixTerm right) throws IOException {
		if (argumentsL.arePresent()) {
			writeLeftArguments(w, UNSAFE_LEFT_MASK & flags, left);
			w.ensureSpace();
		} else if ((flags & UNSAFE_LEFT_PAREN) != 0
				&& (DCX.isAnyPrefixOperatorLike("(", spelling))) {
			w.ensureSpace();
		}
		w.write(spelling);
		if (argumentsR.arePresent()) {
			w.ensureSpace();
			writeRightArguments(w, UNSAFE_RIGHT_MASK & flags, right);
		} else if ((flags & UNSAFE_RIGHT_PAREN) != 0
				&& (DCX.isAnySuffixOperatorLike(")", spelling))) {
			w.ensureSpace();
		}
	}

	@Override
	public void write(MizWriter w, int flags, InfixTerm left, InfixTerm right) throws IOException {
		if ((flags & UNSAFE_LEFT_WEAK_FUNCTOR) != 0
				&& (argumentsL.arePresent() || FRM.canFunctorTake(nr, 1, argumentsR.count()))
				&& FRM.canFunctorTake(left.nr, left.argumentsL.count(), argumentsL.count())
				|| (flags & UNSAFE_RIGHT_WEAK_FUNCTOR) != 0
						&& (argumentsR.arePresent() || FRM.canFunctorTake(nr, argumentsL.count(), 1))
						&& FRM.canFunctorTake(right.nr, argumentsR.count(), right.argumentsR.count())) {
			w.write("(");
			writeIt(w, UNSAFE_LEFT_PAREN | UNSAFE_RIGHT_PAREN, null, null);
			w.write(")");
		} else {
			writeIt(w, flags, left, right);
		}
	}

	@Override
	public void write(MizWriter w, int flags) throws IOException {
		writeIt(w, flags, null, null);
	}

}
